//
//   Copyright 2013 Pixar
//
//   Licensed under the terms set forth in the LICENSE.txt file available at
//   https://opensubdiv.org/license.
//

#include "./version.h"

namespace OpenSubdiv {
namespace OPENSUBDIV_VERSION {


} // end namespace OPENSUBDIV_VERSION
} // end namespace OpenSubdiv
